
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity hacm is
    port 
    (
        clk_i     : in  std_logic;
        rst_i     : in  std_logic;
        
        --the following inputs are synchronous to clk_o
        f_or_s_i  : in  std_logic;
       

        lock_o    : out std_logic;                  
        clk_o     : out std_logic;

        -- configuration interface
        reg_clk   : in std_logic;
        reg_rden  : in std_logic;
        reg_wren  : in std_logic;
        reg_wstrb : in std_logic_vector(3 downto 0);
        reg_rd_data : out std_logic_vector(31 downto 0);
        reg_wr_data : in std_logic_vector(31 downto 0);
        reg_addr    : in std_logic_vector(31 downto 0)
    );

end hacm;

architecture Behavioral of hacm is

component hacm_control
    port 
    (
        clk_i       : in  std_logic;
        rst_i     : in  std_logic;
        
        f_or_s_i  : in  std_logic;
        ts_i      : in  std_logic_vector(7 downto 0);
        tf_i      : in  std_logic_vector(7 downto 0);
        to_i      : in  std_logic_vector(7 downto 0);
        tsw_i     : in  std_logic_vector(7 downto 0);       
        
        mode_o    : out std_logic;
        stretch_o : out std_logic_vector(1 downto 0);   
        sel_o     : out std_logic
    );
end component;

component hacm_core
    port 
    (
        clk_i       : in  std_logic;
        rst_i       : in std_logic;
        mode_i    : in  std_logic; --0=CS; 1=CM
        stretch_i : in  std_logic_vector(1 downto 0);   
        sel_i     : in  std_logic; --0=PC; 1=CO
        lock_o    : out std_logic;                  
        clk_o     : out std_logic;
        --dynamic reconfiguration interface
        drp_addr  : in std_logic_vector(6 downto 0);
        drp_din   : in std_logic_vector(15 downto 0);
        drp_dout  : out std_logic_vector(15 downto 0);
        drp_we    : in std_logic;
        drp_en    : in std_logic;
        drp_rdy   : out std_logic;
        drp_clk   : in std_logic;
        --phase shift interface
        ps_en     : in std_logic;
        ps_incdec : in std_logic;
        ps_done   : out std_logic;
        ps_clk    : in std_logic
    );
end component;

signal ts_i      : std_logic_vector(7 downto 0);
signal tf_i      : std_logic_vector(7 downto 0);
signal to_i      : std_logic_vector(7 downto 0);
signal tsw_i     : std_logic_vector(7 downto 0);

signal reg_index : integer;

signal hacm_mode : std_logic;
signal hacm_stretch : std_logic_vector(1 downto 0);
signal hacm_sel : std_logic;
signal hacm_rst : std_logic;
signal hacm_lock : std_logic;

signal clk_internal : std_logic;

signal drp_we : std_logic;
signal drp_en : std_logic;
signal drp_dout : std_logic_vector(15 downto 0);
signal drp_rdy : std_logic;
signal drp_busy : std_logic;

signal ps_en : std_logic;
signal ps_incdec : std_logic;
signal ps_done : std_logic;
signal ps_busy : std_logic;

begin

reg_index <= to_integer(unsigned(reg_addr(9 downto 2)));

drp_en <= (reg_wren or reg_rden) and reg_addr(9) and (not drp_busy);
drp_we <= reg_wren and reg_addr(9);

ps_en <= (reg_wren and (not ps_busy)) when (reg_index = 8) else '0';
ps_incdec <= ps_en and reg_wr_data(0);

--write registers
process(reg_clk) begin
    if(reg_clk'event and reg_clk = '1') then
        if(reg_wren = '1') then
            case reg_index is
                when 1 => 
                    ts_i <= reg_wr_data(7 downto 0);
                    tf_i <= reg_wr_data(15 downto 8);
                    to_i <= reg_wr_data(23 downto 16);
                    tsw_i <= reg_wr_data(31 downto 24);
                when 0 =>
                    hacm_rst <= reg_wr_data(2);
                when others => --do nothing
            end case;
        end if;
    end if;
end process;

--read registers
process(reg_clk) begin
    if(reg_clk'event and reg_clk = '1') then
        if(reg_rden = '1') then
            if(reg_index < 16) then
                case reg_index is
                    when 0 => 
                        reg_rd_data <= (2 => hacm_rst, 1 => hacm_mode, 0 => hacm_lock, others => '0');
                    when 1 =>
                        reg_rd_data <= tsw_i & to_i & tf_i & ts_i;
                    when 8 =>
                        reg_rd_data <= (1 => ps_busy, others => '0');
                    when others =>
                        reg_rd_data <= (others => '0');
                end case;
            else
                reg_rd_data(15 downto 0) <= drp_dout;
                reg_rd_data(31 downto 16) <= (16 => drp_busy, others => '0');
            end if;
        end if;
    end if;
end process;

-- maintain ps_busy and drp_busy
process(reg_clk) begin
    if(reg_clk'event and reg_clk = '1') then
        if(rst_i = '1') then
            ps_busy <= '0';
            drp_busy <= '0';
        else
            if(ps_en = '1') then
                ps_busy <= '1';
            elsif(ps_done = '1') then
               ps_busy <= '0';
            end if;
            if(drp_en = '1') then
                drp_busy <= '1';
            elsif(drp_rdy = '1') then
                drp_busy <= '0';
            end if;
        end if;
    end if;
end process;

control: hacm_control port map(
    clk_i => clk_internal,
    rst_i => hacm_rst or rst_i,
    f_or_s_i => f_or_s_i,
    ts_i => ts_i,
    tf_i => tf_i,
    to_i => to_i,
    tsw_i => tsw_i,
    
    mode_o => hacm_mode,
    stretch_o => hacm_stretch,
    sel_o => hacm_sel
);

core: hacm_core port map(
    clk_i => clk_i,
    rst_i => hacm_rst or rst_i,
    mode_i => hacm_mode,
    stretch_i => hacm_stretch,
    sel_i => hacm_sel,
    lock_o => hacm_lock,
    clk_o => clk_internal,
    --dynamic reconfiguration interface
    drp_addr => reg_addr(8 downto 2),
    drp_din => reg_wr_data(15 downto 0),
    drp_dout => drp_dout,
    drp_we => drp_we,
    drp_en => drp_en,
    drp_rdy => drp_rdy,
    drp_clk => reg_clk,
    --phase shift interface
    ps_en => ps_en,
    ps_incdec => ps_incdec,
    ps_done => ps_done,
    ps_clk => reg_clk
);

clk_o <= clk_internal;
lock_o <= hacm_lock;


end Behavioral;

